#!/usr/bin/perl -Tw
# autodns 1.0.0
# Copyright 1999-2006 Project Purple. Written by Jonathan McDowell
# See ACKNOWLEDGEMENTS file for full details of contributors.
# http://www.earth.li/projectpurple/progs/autodns.html
# Released under the GPL.
#

use strict;
use Date::Parse;
use Fcntl qw(:flock);
use File::Path;
use File::Temp qw(tempfile);
use IPC::Open3;
use MIME::Parser;

$ENV{'PATH'} = "/usr/local/bin:/usr/bin:/bin:/usr/sbin";

my ($from, $subject, $gpguser, $gpggood, $priv);
my ($user, $server, $inprocess, $delcount, $addcount);
my ($domain, @MAIL, @GPGERROR, @COMMANDS, %zones, $VERSION);

use vars qw($me $ccreply $conffile $domainlistroot @cfgfiles $usersfile
	$lockfile $reload_command $expiry $zonefiledir);

$VERSION = "1.0.0";

#
# Load our config
#
my $file = '/etc/bind/autodns.conf';
unless (my $ret = do $file) {
	warn "Couldn't parse $file\n" if $@;
	warn "Couldn't do $file\n" unless defined $ret;
	warn "Couldn't run $file\n" unless $ret;
	die "Problem reading config file!\n";
}

###
### There should be no need to edit anything below (unless you're not
### using BIND). This statement might even be true now - let me know if not.
###

#
# Try to figure out what zones we currently know about by parsing config
# files. Sets the item in %zones to 1 for each zone it finds.
#
# Call with the name of a config file to read:
#
# getzones("/etc/named.conf");
#
sub getzones($) {
	my $namedfile = shift;

	open (NAMEDCONF, "< $namedfile") or
		fatalerror("Can't open $namedfile");

	while (<NAMEDCONF>) {
		if (/^\s*zone\s*"([^"]+)"/) {
			$zones{$1}=1;
		}
	}

	close NAMEDCONF;
}

#
# Check that a domain is only made up of valid characters.
#
# These are: a-z, 0-9, - or .
#
sub valid_domain($) {
	my $domain = shift;
	$domain = lc $domain;

	if ($domain =~ /^(?:[a-z0-9-]+\.)+[a-z]{2,6}$/) {
		return 1;
	} elsif ($domain =~ /^(?:[0-9\/-]+\.)+in-addr.arpa$/) {
		return 1;
	} else {
		return 0;
	}
}

#
# Deal with a fatal error by printing an error message, closing the pipe to
# sendmail and exiting.
#
# fatalerror("I'm melting!");
#
sub fatalerror($) {
	my $message = shift;

	print REPLY $message;
	close(REPLY);

	flock(LOCKFILE, LOCK_UN);
	close(LOCKFILE);
	unlink($lockfile);

	exit;
}

#
# Get user details from usersfile based on a PGP ID.
#
# A users entry looks like:
#
# <username>:<keyid>:<privilege level>:<master server ip>
#
# Priviledge level is not currently used.
#
# ($user, $priv, $server) = getuserinfo("5B430367");
#
sub getuserinfo($) {
	my $gpguser = shift;
	my ($user, $privilege, $server);

	open (CONFIGFILE, "< $usersfile") or
		fatalerror("Couldn't open user configuration file.");

	foreach (<CONFIGFILE>) {
		if (/^([^#.]+):$gpguser:(\d+):(.+)$/) {
			$user = $1;
			$privilege = $2;
			$server = $3;
			chomp $user;
			chomp $privilege;
			chomp $server;
	
			if ($user !~ /^.+$/) {
				close(CONFIGFILE);
				fatalerror("Error in user configuration ".
						"file: Can't get username.\n");
			}

			if ($server !~ /^(\d{1,3}\.){3}\d{1,3}$/) {
				$server =~ s/\d\.]//g;
				close(CONFIGFILE); 
				fatalerror("Error in user configuration ".
					"file: Invalid primary server IP ".
					"address ($server)\n");
				exit;
			} 
		}
	} 
	close(CONFIGFILE);

	if ($user =~ /^$/) {
		fatalerror("User not found.\n");
	}

	return ($user, $privilege, $server);
}

#
# Add a new AutoDNS user.
#
# addautodnsuser($username, $keyid, $priv, $masterip);
# <username>:<keyid>:<privilege level>:<master server ip>
#
sub addautodnsuser($$$$) {
	my $username = shift;
	my $keyid = shift;
	my $priv = shift;
	my $masterip = shift;

	# Create domains file for the user.
	open (DOMAINLIST, ">>$domainlistroot$username") or
			fatalerror("Couldn't create domains file.\n");
	close DOMAINLIST;

	# Make the directory for the zone files.
	my @dirs = mkpath("$zonefiledir/$username", 0, 0775);
	fatalerror("Couldn't create zone file directory.\n")
			if scalar(@dirs) == 0;

	# Actually add them to the users file.
	open(USERFILE, ">> $usersfile") or
		fatalerror("Couldn't open user configuration file.");
	print USERFILE "$username:$keyid:$priv:$masterip\n";
	close(USERFILE);
}

$delcount = $addcount = $inprocess = 0;

# Read in the mail from stdin.
@MAIL = <>;

$subject = "Reply from AutoDNS";
# Now lets try to find out who it's from.
foreach (@MAIL) {
	if (/^$/) { last; }
	if (/^From: (.*)/i) { $from = $1; chomp $from;}
	if (/^Subject:\s+(re:)?(.*)$/i) { $subject = "Re: ".$2 if ($2);}
}

if ((! defined($from)) || $from =~ /^$/ ) {
	die "Couldn't find a from address.";
} elsif ($from =~ /mailer-daemon@/i) {
	die "From address is mailer-daemon, ignoring.";
}

if (! defined($subject)) { $subject = "Reply from AutoDNS"; };

# We've got a from address. Start a reply.

open(REPLY, "|sendmail -t -oem -oi") or die "Couldn't spawn sendmail";

print REPLY "From: $me\n";
print REPLY "To: $from\n";
#
# Check to see if our CC address is the same as the from address and if so
# don't CC.
#
if ($from ne $ccreply) {
	print REPLY "Cc: $ccreply\n";
}
print REPLY <<EOF;
Subject: $subject

AutoDNS $VERSION
Copyright 1999-2006 Project Purple. Written by Jonathan McDowell.
Released under the GPL.

EOF

#
# Throw the mail at MIME::Parser and see if it accepts it.
#
my $parser = new MIME::Parser;
$parser->output_to_core(1); # No temporary files
my $entity = $parser->parse_data(\@MAIL);

#
# Make sure locale is set to C so we get messages in English as we expect.
#
$ENV{'LC_ALL'} = "C";

if ($entity->parts) {
	# MIME

	my ($got_sig, $got_text) = (0, 0);
	my ($sig_fh, $sig_name) = tempfile();
	my ($text_fh, $text_name) = tempfile();

	foreach my $subent ($entity->parts) {
		if ($subent->effective_type eq "text/plain") {
			@COMMANDS = split /\n/,$subent->bodyhandle->as_string;

			my $str = $subent->as_string;
			$str =~ s/=\n$//;
			$str =~ s/\n/\r\n/g;
			print $text_fh $str;
			close($text_fh);
			$got_text++;
		} elsif ($subent->effective_type eq
				"application/pgp-signature") {
			print $sig_fh $subent->as_string;
			close($sig_fh);
			$got_sig++;
	 	} elsif ($subent->effective_type eq "multipart/mixed") {
			my $str = $subent->as_string;
			print $text_fh $str;
			close($text_fh);
			$got_text++;
	
			foreach my $mixent ($subent->parts) {
				if ($mixent->effective_type eq "text/plain") {
					push @COMMANDS, (split /\n/,
						$mixent->bodyhandle->as_string);
				}
				if ($mixent->effective_type eq
						"application/pgp-keys") {
					push @COMMANDS, (split /\n/,
						$mixent->bodyhandle->as_string);
				}
			}
		}
	}

	if ($got_sig && $got_text) {
		my $pid = open3(\*GPGIN, \*GPGOUT, \*GPGERR,
			"gpg --batch --verify ".
			$sig_name." ".$text_name);

		close GPGIN;

		@GPGERROR = <GPGERR>;
		my @GPGOUTPUT = <GPGOUT>;
		close GPGERR;
		close GPGOUT;
		waitpid $pid, 0;

		unlink($text_name);
		unlink($sig_name);
	}
} else {
	# Clear text.

	my $pid = open3(\*GPGIN, \*GPGOUT, \*GPGERR, "gpg --batch --verify");

	# Feed it the mail.
	print GPGIN $entity->bodyhandle->as_string;
	close GPGIN;

	# And grab what it has to say.
	@GPGERROR = <GPGERR>;
	@COMMANDS = <GPGOUT>;
	@COMMANDS = split /\n/,$entity->bodyhandle->as_string;
	close GPGERR;
	close GPGOUT;
	waitpid $pid, 0;
}

# Check who it's from and if the signature was a good one.
$gpggood=1;
my $sigtime = 0;
foreach (@GPGERROR) {
	chomp;
	if (/Signature made (.*) using.*ID (.*)$/) {
		$sigtime = str2time($1);
		$gpguser = $2;
	} elsif (/error/) {
		$gpggood = 0;
		print REPLY "Some errors ocurred\n";
	} elsif (/BAD signature/) {
		$gpggood = 0;
		print REPLY "BAD signature!\n";
	} elsif (/public key not found/) {
		$gpggood = 0;
		print REPLY "Public Key not found\n";
	}
}

# If we have an empty user it probably wasn't signed.
if (! $gpguser) {
	print REPLY "Message appears not to be GPG signed.\n";
	close REPLY;
	exit;
}

# Check the signature we got was ok.
if ($gpggood) {
	print REPLY "Good GPG signature found. ($gpguser)\n";
} else {
	print REPLY "Bad GPG signature!\n";
	close REPLY;
	exit;
}

# Check if the signature is outside our acceptable range.
if (!defined($sigtime)) {
	print REPLY "Couldn't parse signature time.\n";
	close REPLY;
	exit;
} elsif ($sigtime > (time + $expiry)) {
	print REPLY "Signature too far into the future.\n";
	close REPLY;
	exit;
} elsif ($sigtime < (time - $expiry)) {
	print REPLY "Signature too far into the past.\n";
	close REPLY;
	exit;
}

# Now let's check if we know this person.
($user, $priv, $server) = getuserinfo($gpguser);

if (! defined($user) || ! $user) {
	print REPLY "Unknown user.\n";
	close REPLY;
	exit;
}

print REPLY "Got user '$user'\n";

# Right. We know this is a valid user. Get a lock to ensure we have exclusive
# access to the configs from here on in.
open (LOCKFILE,">$lockfile") ||
	 fatalerror("Couldn't open lock file\n");
fatalerror("Couldn't get lock\n") unless(flock(LOCKFILE,LOCK_EX));

# Ok, now we should figure out what domains we already know about.
foreach my $cfgfile (@cfgfiles) {
	getzones($cfgfile);
}

# Force existance of the $domainlistroot$user file
if (! -e $domainlistroot.$user) {
	open (DOMAINLIST, ">>$domainlistroot$user") or
			fatalerror("Couldn't create domains file.\n");
	close DOMAINLIST;
}

foreach (@COMMANDS) {
	# Remove trailing CRs and leading/trailing whitespace
	chomp;
	s/\r//;
	s/^\s*//;
	s/\s*$//;

	if ($inprocess) {
		print REPLY ">>>$_\n";
	}

	if (/^$/) {
		#
		# Empty line, so ignore it.
		# 
	} elsif (/^END$/) {
		$inprocess = 0;
	} elsif (/^BEGIN$/) {
		$inprocess = 1;
	} elsif ($inprocess && /^ADD\s+(.*)$/) {
		$domain = $1;

		# Convert domain to lower case.
		$domain =~ tr/[A-Z]/[a-z]/;
		if (! valid_domain($domain)) {
			$domain =~ s/[-a-z0-9.]//g;
			print REPLY "Invalid character(s) in domain name:",
					" $domain\n";
		} elsif (defined($zones{$domain}) && $zones{$domain}) {
			print REPLY "We already secondary $domain\n";
		} else {
			print REPLY "Adding domain $domain\n";
			$zones{$domain} = 1;

			my $df = $domain;
			$df =~ tr,/,:,;

			open (DOMAINSFILE, ">>$conffile");
			print DOMAINSFILE "
### Domain added for '$user'

zone \"$domain\" {
	type slave;
	masters { $server; };
	file \"secondary/$user/$df\";
	allow-transfer { none; };
	allow-query { any; };
};\n";
			close DOMAINSFILE;

			open (DOMAINLIST, ">>$domainlistroot$user") or
				fatalerror("Couldn't open file.\n");
			print DOMAINLIST "$domain\n";
			close DOMAINLIST;
			$addcount++;
		}
	} elsif ($inprocess && /^DEL\s(.*)$/) {
		$domain = $1;

		# Convert domain to lower case.
		$domain =~ tr/[A-Z]/[a-z]/;
		if (!valid_domain($domain)) {
			$domain =~ s/[-a-z0-9.]//g;
			print REPLY "Invalid character(s) in domain name:",
					" $domain\n";
		} elsif (!defined($zones{$domain}) || !$zones{$domain}) {
				print REPLY "$domain does not exist!\n";
		} else {
			print REPLY "Deleting domain $domain\n";
			my (@newcfg, $found);

			open (DOMAINLIST, "<$domainlistroot$user") or
				fatalerror("Couldn't open file ".
						$domainlistroot.$user.
						" for reading: $!.\n");
			my @cfg = <DOMAINLIST>;
			close(DOMAINLIST);
			@newcfg = grep { ! /^$domain$/ } @cfg;
			if (scalar @cfg == scalar @newcfg) {
				print REPLY "Didn't find $domain in ",
						"$domainlistroot$user!\n";
				print REPLY "You are only allowed to delete",
					" your own domains that exist.\n";
				next;
			}

			open (DOMAINLIST, ">$domainlistroot$user") or 
				fatalerror("Couldn't open file ".
						$domainlistroot.$user.
						" for writing: $!.\n");
			print DOMAINLIST @newcfg;
			close DOMAINLIST;

			$found = 0;
			@newcfg = ();
			open (DOMAINSFILE, "<$conffile") or
				fatalerror("Couldn't open file $conffile for".
						" reading: $!\n");
			{
			local $/ = ''; # eat whole paragraphs
			while (<DOMAINSFILE>) {
				unless (/^\s*zone\s+"$domain"/) {
					push @newcfg, $_;
				} else {
					$found = 1;
					if ($newcfg[-1] =~ /^###/) {
						# remove comment and \n
						pop @newcfg;
					}
				}
			}
			} # end of paragraph eating

			if (!$found) {
				print REPLY "Didn't find $domain in",
						" $conffile!\n";
				next;
			}

			open (DOMAINSFILE, ">$conffile") or
				fatalerror("Couldn't open $conffile for".
						" writing: $!\n");
			print DOMAINSFILE @newcfg;
			close DOMAINSFILE;
			$delcount++;
			$zones{$domain} = 0;
		}
	} elsif ($inprocess && /^LIST$/) {
		print REPLY "Listing domains for user $user\n";
		print REPLY "------\n";
		if (open (DOMAINLIST, "<$domainlistroot$user")) {
			my $count = 0;
			while (<DOMAINLIST>) {
				$count++;
				print REPLY;
			}
			close (DOMAINLIST);
			print REPLY "------\n";
			print REPLY "Total of $count domains.\n";
		} else {
			print REPLY "Couldn't open $domainlistroot$user: $!\n";
		}
	} elsif ($inprocess && /^MASTER\s(.*)$/) {
		if (($priv & 1) != 1) {
			print REPLY "You're not authorised to use the MASTER ",
				"command.\n";
		} elsif ($1 =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/) {
			$server = $1;
			print REPLY "Set master IP address to $1\n";
		} else {
			print REPLY "$1 doesn't look like a valid IPv4 ",
				"address to me.\n";
		}
	} elsif ($inprocess && /^ADDUSER\s(.*)$/) {
		if (($priv & 2) != 2) {
			print REPLY "You're not authorised to use the ",
				"ADDUSER command.\n";
		} elsif ($1 =~ /^([a-z0-9]+) ([A-Fa-f0-9]{8}) (\d+) (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/) {
			addautodnsuser($1, $2, $3, $4);

			print REPLY "Attempting to import new key:\n";

			# Feed our command mail to GPG so we can pull the
			# (hopefully included) new GPG key out from it.
			my $pid = open3(\*GPGIN, \*GPGOUT, \*GPGERR,
					"gpg --batch --import");

			# Feed it the mail.
			print GPGIN join("\n", @COMMANDS);
			close GPGIN;

			# And grab what it has to say.
			@GPGERROR = <GPGERR>;
			my @GPGOUTPUT = <GPGOUT>;
			close GPGERR;
			close GPGOUT;
			waitpid $pid, 0;

			print REPLY @GPGERROR;
		} else {
			print REPLY "ADDUSER parameter error.\n";
		}
	} elsif ($inprocess && /^HELP$/) {
		print REPLY <<EOF;
In order to use the service, you will need to send GPG signed messages.
The format of the text in these messages is important, as they represent
commands to autodns. Commands are formatted one per line, and enclosed
by "BEGIN" and "END" commands (without the quotes).

Current valid commands are:

BEGIN - begin processing.
END - end processing.
HELP - display this message.
LIST - show all the zones currently held by you.
ADD <domain> - adds the domain <domain> for processing.
DEL <domain> - removes the domain <domain> if you own it.
EOF
		if (($priv & 1) == 1) {
			print REPLY "MASTER <ip address> - set the nameserver".
			" we should slave off for subsequent ADD\ncommands.\n";
		}
		if (($priv & 2) == 2) {
			print REPLY "ADDUSER <username> <keyid> <privilege> ",
				"<masterip> - add a new user. Imports any key",
				"\nattached to the message into the keyring.\n";
		}
	} elsif ($inprocess) {
		print REPLY "Unknown command!\n";
	}
}
flock(LOCKFILE, LOCK_UN);
close(LOCKFILE);
unlink($lockfile);

print REPLY "Added $addcount domains.\n" if $addcount;
print REPLY "Removed $delcount domains.\n" if $delcount;
if ($addcount || $delcount) {
	print REPLY "Reloading nameserver config.\n";
	print REPLY `$reload_command`;
}
close REPLY;

exit 0;
